/*
 * mjpegserver - simulates multiple network camera vendor/models
 * Copyright (C) 2005  Christopher M. Price
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <errno.h>
#include <signal.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>

#include "netcams.h"

/* defaults */
#define PORT 8080
#define PATH "images"

struct image_t {
    char *ptr;
    size_t size;
};

/* global options */
struct opts_t {
    int type;
    int port;
    char *path;
} opts;

void error(char *);
void help(void);
void child_run(int);
void create_child(int);
void server(void);
void push_image(int, char *, struct image_t *);
void push_stream(int);

int main(int argc, char *argv[])
{
    int opt, i = 0;

    opts.type = -1;
    opts.port = PORT;
    opts.path = "images";

    while ((opt = getopt(argc, argv, "d:hp:t:")) != -1) {
	switch (opt) {
	case 'd':
	    opts.path = optarg;
	    break;
	case 'p':
	    opts.port = atoi(optarg);
	    break;
	case 't':
	    opts.type = atoi(optarg);
	    break;
	case 'h':
	    help();
	    break;
	default:
	    error("getopt() failed");
	}
    }
    do {
	if (netcam[i].id == opts.type)
	    break;
	i++;
    } while (netcam[i].id != -1);
    if (netcam[i].id == -1) {
	fprintf(stderr,
		"mjpegserver: camera type not specified/supported\n");
	exit(-1);
    }
    fprintf(stderr, "mjpegserver: id: %d, type: %s\n",
	    netcam[i].id, netcam[i].name);
    fprintf(stderr, "mjpegserver: port: %d\n", opts.port);
    fprintf(stderr, "mjpegserver: path: %s\n", opts.path);

    server();

    return 0;
}

void error(char *msg)
{
    if (errno != 0)
	perror(msg);
    else
	fprintf(stderr, "%s\n", msg);
    exit(-1);
}

void help(void)
{
    int i = 0;
    fprintf(stderr,
	    "Usage: mjpegserver [OPTION]\n"
	    " -h                help\n"
	    " -d                image directory\n"
	    " -p<PORT>          port number to listen on\n"
	    " -t<ID>            netcam type (see below)\n" "\n");
    fprintf(stderr, "ID VENDOR/MODELS\n");
    do {
	fprintf(stderr, "%2d %s\n", netcam[i].id, netcam[i].name);
	i++;
    } while (netcam[i].id != -1);
    exit(0);
}

void create_child(int ssd)
{
    int csd;
    struct sockaddr_in cad;
    int len;

    switch (fork()) {
    case -1:
	error("fork()");
	break;
    case 0:
	fprintf(stderr, "mjpegserver: process %d started\n", getpid());
	len = sizeof(cad);
	if ((csd = accept(ssd, (struct sockaddr *) &cad, &len)) < 0)
	    error("accept()");
	close(ssd);
	push_stream(csd);
	break;
    default:
	break;
    }
}

#define MAX_CHLD 4

void server(void)
{
    struct sockaddr_in sad;
    int ssd;
    int optval, i;

    if ((ssd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	error("socket() failed");

    optval = 1;
    if ((setsockopt(ssd, SOL_SOCKET, SO_REUSEADDR,
		    (int *) &optval, sizeof(optval))) < 0)
	error("setsockopt() failed");

    memset(&sad, 0, sizeof(sad));
    sad.sin_family = AF_INET;
    sad.sin_addr.s_addr = htonl(INADDR_ANY);
    sad.sin_port = htons(opts.port);

    if (bind(ssd, (struct sockaddr *) &sad, sizeof(sad)) < 0)
	error("bind() failed");

    if (listen(ssd, 8) < 0)
	error("listen() failed");

    for (i = 0; i < MAX_CHLD; i++)
	create_child(ssd);

    while (1) {
	fprintf(stderr, "mjpegserver: process %d terminated\n", wait(0));
	create_child(ssd);
    }
}

void push_stream(int csd)
{
    struct dirent *direntp;
    DIR *dirp;
    struct image_t image;

    image.ptr = NULL;
    image.size = 0;
    if ((dirp = opendir(opts.path)) == NULL)
	error("opendir() failed");
    else {
	chdir(opts.path);
	if (send(csd, netcam[opts.type].header_http,
		 strlen(netcam[opts.type].header_http), 0) < 0)
	    error("header_http send() failed");
	do {
again:
	    while ((direntp = readdir(dirp)) == NULL) {
		rewinddir(dirp);
	    }
	    if (strcmp(direntp->d_name, ".") == 0 ||
		strcmp(direntp->d_name, "..") == 0)
		goto again;

	    push_image(csd, direntp->d_name, &image);

	    sleep(1);
	} while (1);
	free(image.ptr);
	closedir(dirp);
    }
}

void push_image(int csd, char *fname, struct image_t *image)
{
    char buf[1024];
    struct stat imagefd_stat;
    int imagefd;

    if ((imagefd = open(fname, 0)) < 0)
	error("open() failed");
    if (fstat(imagefd, &imagefd_stat) < 0)
	error("fstat() failed");
    if (image->size < imagefd_stat.st_size) {
	if ((image->ptr =
	     realloc(image->ptr, imagefd_stat.st_size)) == NULL)
	    error("realloc() failed");
    }
    if (read(imagefd, image->ptr, imagefd_stat.st_size) < 0)
	error("read() failed");
    image->size = imagefd_stat.st_size;
    close(imagefd);

    sprintf(buf, netcam[opts.type].header_image, image->size);
    if (send(csd, buf, strlen(buf), 0) < 0)
	error("header_image send() failed");
    if (send(csd, image->ptr, image->size, 0) < 0)
	error("image send() failed");
    if (netcam[opts.type].header_image_sep != NULL)
	if (send(csd, netcam[opts.type].header_image_sep,
		 strlen(netcam[opts.type].header_image_sep), 0) < 0)
	    error("header_image send() failed");
}
